#include "adjustnonlinear.h"

const double PI = 3.14159265358979;
const double PI2 = 2*PI;

double AdjustNonLinear::adjust(int npts, p3d *mpts, p3d *cpts, p3d *&tpts, params *pa, double &rms)
{
	double* tr;						//array of trig shortcuts so that cos() & sin() only need to be called once
	tr = new double[6];

	// make an initial guess
	initialGuess( pa, tpts, mpts, npts, tr );

	// iterate through the least squares adjustment
	rms = 0;
	int i;
	for( i = 0; i < 200; i++ )
	{
		double temprms;
		temprms = iterate( npts, mpts, cpts, tpts, pa, tr );

		// force the angles to be between -2pi and 2pi
		pa->ax = angleRangeCorrect(pa->ax);
		pa->ay = angleRangeCorrect(pa->ay);
		pa->az = angleRangeCorrect(pa->az);

		// escape the loop if the difference between the current iteration
		// and the previous iteration is small enough
		if( qAbs(temprms - rms) < 0.000001 )
		{
			rms = temprms;
			break;
		}
		rms = temprms;
	}

	// compute the rotation matrix
	RotMat = new double*[3];
	for( int i = 0; i < 3; i++ )
		RotMat[i] = new double[3];
	RotMat[0][0] = tr[1]*tr[2];
	RotMat[0][1] = tr[0]*tr[5]+tr[3]*tr[4]*tr[2];
	RotMat[0][2] = tr[3]*tr[5]-tr[0]*tr[4]*tr[2];
	RotMat[1][0] = -tr[1]*tr[5];
	RotMat[1][1] = tr[0]*tr[2]-tr[3]*tr[4]*tr[5];
	RotMat[1][2] = tr[3]*tr[2]+tr[0]*tr[4]*tr[5];
	RotMat[2][0] = tr[4];
	RotMat[2][1] = -tr[3]*tr[1];
	RotMat[2][2] = tr[0]*tr[1];

	return i;
}

int AdjustNonLinear::initialGuess( params* &pa, p3d* &tpts, p3d* mpts, int n, double* &tr )
{
	pa->s = 1;
	pa->ax = pa->ay = pa->az = 0;
	pa->tx = pa->ty = pa->tz = 0;

	tr[0] = cos(pa->ax); tr[1] = cos(pa->ay); tr[2] = cos(pa->az);
	tr[3] = sin(pa->ax); tr[4] = sin(pa->ay); tr[5] = sin(pa->az);

	delete[] tpts;
	tpts = new p3d[n];
	transform( n, mpts, tpts, pa, tr );

	return 0;
}

void AdjustNonLinear::transform( int n, p3d* mpts, p3d* &tpts, params* pa, double* tr )
{
	for( int i = 0; i < n; i++ )
	{
		tpts[i].x = pa->s * ( ( tr[1]*tr[2])*mpts[i].x + (tr[0]*tr[5]+tr[3]*tr[4]*tr[2])*mpts[i].y + (tr[3]*tr[5]-tr[0]*tr[4]*tr[2])*mpts[i].z ) + pa->tx;
		tpts[i].y = pa->s * ( (-tr[1]*tr[5])*mpts[i].x + (tr[0]*tr[2]-tr[3]*tr[4]*tr[5])*mpts[i].y + (tr[3]*tr[2]+tr[0]*tr[4]*tr[5])*mpts[i].z ) + pa->ty;
		tpts[i].z = pa->s * ( ( 1*tr[4]  )*mpts[i].x + (  -tr[3]*tr[1]      )*mpts[i].y + (   tr[0]*tr[1]      )*mpts[i].z ) + pa->tz;
	}
}

double AdjustNonLinear::iterate( int npts, p3d* mpts, p3d* cpts, p3d* &tpts, params* &pa, double* &tr )
{
	double rms;						//root mean square error
	double** jmat;					//observation equation matrix (3 rows per point)... aka jacobian matrix[r][c]
	double** nmat;					//normal matrix
	double** kmat;					//K matrix (control coordinates - transformed coordinates)
	double** corr;					//guess correction vector

	//create the J matrix. syntax: jmat[row][col]
	// has 7 columns for the 7-element 3D Conformal Coordinate Transformation
	jmat = new double*[npts*3];				//declare 3 rows per point
	for( int r = 0; r < npts*3; r++ )
		jmat[r] = new double[7];			//declare 7 columns per row

	// populate the J matrix with the linearized observation equations, which are
	// partial derivatives of the following rotation matrix:
	//  [  cos(Ry)cos(Rz)  cos(Rx)sin(Rz)+sin(Rx)sin(Ry)cos(Rz)  sin(Rx)sin(Rz)-cos(Rx)sin(Ry)cos(Rz) ]
	//  [ -cos(Ry)sin(Rz)  cos(Rx)cos(Rz)-sin(Rx)sin(Ry)sin(Rz)  sin(Rx)cos(Rz)+cos(Rx)sin(Ry)sin(Rz) ]
	//  [  sin(Ry)         -sin(Rx)cos(Ry)                       cos(Rx)cos(Ry)                       ]
	for( int r = 0; r < npts*3; r+=3 )
	{
		int p = floor(r/3);	//each point contributes 3 rows to the J matrix
		/* dF/dS */
		jmat[r+0][0] = ( tr[1]*tr[2])*mpts[p].x + ( tr[0]*tr[5]+tr[3]*tr[4]*tr[2])*mpts[p].y + (tr[3]*tr[5]-tr[0]*tr[4]*tr[2])*mpts[p].z;
		jmat[r+1][0] = (-tr[1]*tr[5])*mpts[p].x + ( tr[0]*tr[2]-tr[3]*tr[4]*tr[5])*mpts[p].y + (tr[3]*tr[2]+tr[0]*tr[4]*tr[5])*mpts[p].z;
		jmat[r+2][0] = ( tr[4]*1.0  )*mpts[p].x + (-tr[3]*tr[1]         )*mpts[p].y + (tr[0]*tr[1]         )*mpts[p].z;
		/* dF/dRx */
		jmat[r+0][1] = pa->s * ( (-tr[3]*tr[5]+tr[0]*tr[4]*tr[2])*mpts[p].y + (tr[0]*tr[5]+tr[3]*tr[4]*tr[2])*mpts[p].z );
		jmat[r+1][1] = pa->s * ( (-tr[3]*tr[2]-tr[0]*tr[4]*tr[5])*mpts[p].y + (tr[0]*tr[2]-tr[3]*tr[4]*tr[5])*mpts[p].z );
		jmat[r+2][1] = pa->s * ( (-tr[0]*tr[1]         )*mpts[p].y + (-tr[3]*tr[1]        )*mpts[p].z );
		/* dF/dRy */
		jmat[r+0][2] = pa->s * ( (-tr[4]*tr[2])*mpts[p].x + ( tr[3]*tr[1]*tr[2])*mpts[p].y + (-tr[0]*tr[1]*tr[2])*mpts[p].z );
		jmat[r+1][2] = pa->s * ( ( tr[4]*tr[5])*mpts[p].x + (-tr[3]*tr[1]*tr[5])*mpts[p].y + ( tr[0]*tr[1]*tr[5])*mpts[p].z );
		jmat[r+2][2] = pa->s * ( ( tr[1]*1.0  )*mpts[p].x + ( tr[3]*tr[4]   )*mpts[p].y + (-tr[0]*tr[4]   )*mpts[p].z );
		/* dF/dRz */
		jmat[r+0][3] = pa->s * ( (-tr[1]*tr[5])*mpts[p].x + ( tr[0]*tr[2]-tr[3]*tr[4]*tr[5])*mpts[p].y + ( tr[3]*tr[2]+tr[0]*tr[4]*tr[5])*mpts[p].z );
		jmat[r+1][3] = pa->s * ( (-tr[1]*tr[2])*mpts[p].x + (-tr[0]*tr[5]-tr[3]*tr[4]*tr[2])*mpts[p].y + (-tr[3]*tr[5]+tr[0]*tr[4]*tr[2])*mpts[p].z );
		jmat[r+2][3] = 0.0;
		/* dF/dTx */
		jmat[r+0][4] = 1.0;
		jmat[r+1][4] = 0.0;
		jmat[r+2][4] = 0.0;
		/* dF/dTy */
		jmat[r+0][5] = 0.0;
		jmat[r+1][5] = 1.0;
		jmat[r+2][5] = 0.0;
		/* dF/dTz */
		jmat[r+0][6] = 0.0;
		jmat[r+1][6] = 0.0;
		jmat[r+2][6] = 1.0;
	}

	// Create and populate the K matrix where each element equals the control coordinate
	// minus the transformed coordinate. It will be three elements tall for each point.
	kmat = new double*[npts*3];
	for( int k = 0; k < npts*3; k++ )
		kmat[k] = new double[1];		//the K vector is a double** matrix to match the matrix multiplication syntax
	for( int k = 0; k < npts*3; k+=3 )
	{
		int p = floor(k/3);	//each point contributes 3 rows to the K matrix
		//k value = (control) - (transformed coordinate)
		kmat[k][0]   = cpts[p].x - tpts[p].x;
		kmat[k+1][0] = cpts[p].y - tpts[p].y;
		kmat[k+2][0] = cpts[p].z - tpts[p].z;
	}

	// create the normal matrix. syntax: nmat[npts rows][npts columns]
	nmat = new double*[7];
	for( int r = 0; r < 7; r++ )
		nmat[r] = new double[7];
	// populate the normal matrix (Jmat-transpose * Jmat)
	nmat = matrixMultTrans( jmat, npts*3, 7, jmat, npts*3, 7 );
	// overwrite the normal matrix with its inverse
	for( int k = 0; k < 7; k++ )
	{
		for( int j = 0; j < 7; j++ )
			if( j != k )
				nmat[k][j] = nmat[k][j]/nmat[k][k];
		nmat[k][k] = 1.0/nmat[k][k];
		for( int i = 0; i < 7; i++ )
			if( i != k )
			{
				for( int j = 0; j < 7; j++ )
					if( j != k )
						nmat[i][j] = nmat[i][j] - nmat[i][k]*nmat[k][j];
				nmat[i][k] = -nmat[i][k]*nmat[k][k];
			} //end if i<>k
	} //end for k

	// compute corrections
	corr = new double*[7];
	for( int i = 0; i < 7; i++ )
		corr[i] = new double[1];		//the corrections vector is a double** matrix to match the matrix multiplication syntax
	corr = matrixMult( nmat, 7, 7, matrixMultTrans( jmat, npts*3, 7, kmat, npts*3, 1 ), 7, 1 );

	// update parameters with computed corrections
	pa->s  += corr[0][0];
	pa->ax += corr[1][0];
	pa->ay += corr[2][0];
	pa->az += corr[3][0];
	pa->tx += corr[4][0];
	pa->ty += corr[5][0];
	pa->tz += corr[6][0];
	tr[0] = cos(pa->ax); tr[1] = cos(pa->ay); tr[2] = cos(pa->az);
	tr[3] = sin(pa->ax); tr[4] = sin(pa->ay); tr[5] = sin(pa->az);

	// update transformed point coordinates
	transform( npts, mpts, tpts, pa, tr );

	// compute the root mean square error
	rms = 0.0;
	for( int i = 0; i < npts; i++ )
	{
		rms += (cpts[i].x - tpts[i].x)*(cpts[i].x - tpts[i].x) + (cpts[i].y - tpts[i].y)*(cpts[i].y - tpts[i].y) + (cpts[i].z - tpts[i].z)*(cpts[i].z - tpts[i].z);
	}
	rms = sqrt( rms / npts );

	delete [] jmat;
	delete [] nmat;
	delete [] kmat;
	delete [] corr;

	return rms;
}

double AdjustNonLinear::angleRangeCorrect(double a)
{
	while( a >= PI2 )
		a -= PI2;
	while( a <= -PI2 )
		a += PI2;
	return a;
}

double** AdjustNonLinear::matrixMult( double** m1, int nr1, int nc1, double** m2, int nr2, int nc2 )
{
	// verify that the number of columns in m1 equals the number of rows in m2
	if( nc1 != nr2 )
	{
		return 0;
	}

	// create the result matrix, dim nr1 x nc2. syntax: mat[row][col]
	double** mr = new double*[nr1];			//declare nr1 rows
	for( int r = 0; r < nr1; r++ )
	{
		mr[r] = new double[nc2];			//declare nc2 columns
	}
	// matrix multiply m1 x m2
	for( int r = 0; r < nr1; r++ )			//cycle through the rows of m1
	{
		for( int c = 0; c < nc2; c++ )		//cycle through the columns of m2
		{
			mr[r][c] = 0;					//initialize to zero
			for( int i = 0; i < nc1; i++ )	//take the dot product of row r in m1 with column c in m2
			{
				mr[r][c] += m1[r][i] * m2[i][c];
			} //end for dot product
		} //end for each column
	} //end for each row

	return mr;
}

// same as matrixMult but with the first matrix transposed
double** AdjustNonLinear::matrixMultTrans( double** m1, int nr1, int nc1, double** m2, int nr2, int nc2 )
{
	if( nr1 != nr2 )
		return 0;

	// create the result matrix, dim nc1 x nc2. syntax: mat[row][col]
	double** mr = new double*[nc1];
	for( int r = 0; r < nc1; r++ )
		mr[r] = new double[nc2];

	// matrix multiply m1 x m2
	for( int r = 0; r < nc1; r++ )
	{
		for( int c = 0; c < nc2; c++ )
		{
			mr[r][c] = 0;
			for( int i = 0; i < nr1; i++ )
			{
				mr[r][c] += m1[i][r] * m2[i][c];
			} //end for dot product
		} //end for each column
	} //end for each row

	return mr;
}
