#include <QLabel>
#include <QVBoxLayout>
#include <QPushButton>
#include <QFileDialog>
#include <QFile>
#include <QString>
#include <QCoreApplication>
#include <QPlainTextEdit>


#include "mainwindow.h"

const double PI = 3.14159265358979;

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
{
	setWindowFlags( windowFlags() & ~Qt::WindowMaximizeButtonHint );//disable the maximize button

	// initialize pointers and counters to zero
	npts = 0;
	ctrlpts = measpts = transpts = 0;
	trans_params = new params[1];
	trans_params->ax = trans_params->ay = trans_params->az = 0;
	trans_params->tx = trans_params->ty = trans_params->tz = 0;
	trans_params->s = 1;

	// create a very basic layout with just a single label and a button
	// create a minimalistic layout with just a text window and button
	text_ViewLog = new QPlainTextEdit();
		text_ViewLog->setMinimumSize(400,210);								//sets the minimimum (and default) widget size (optional)
		text_ViewLog->setReadOnly(true);									//prevents editing by the user (required when being used as a display widget)
		text_ViewLog->setTabChangesFocus(true);								//allows user to tab-cycle through the objects (Tab does nothing if false and it's ReadOnly)
		text_ViewLog->setLineWrapMode(QPlainTextEdit::NoWrap);				//prevents wordwrap so that each line of text is a unique message (optional)
//		text_ViewLog->setPlainText("Hello");								//sample starter text (optional. note that keeping this blank will clear the display)
//		text_ViewLog->appendHtml("<b>World</b>");							//sample addition of html-formatted text
	pushbutton_Start = new QPushButton(tr("Start"));
		pushbutton_Start->setFixedWidth(60);
	layout_Main = new QVBoxLayout();
		layout_Main->addWidget(text_ViewLog);
		layout_Main->setAlignment(text_ViewLog, Qt::AlignHCenter);
		layout_Main->addWidget(pushbutton_Start);
		layout_Main->setAlignment(pushbutton_Start, Qt::AlignHCenter);
	centralWidget = new QWidget();
		centralWidget->setLayout(layout_Main);
	this->setCentralWidget(centralWidget);

	// weave signals and slots
	connect(pushbutton_Start, SIGNAL(clicked()), this, SLOT(pushbutton_Start_onClicked()));

	// import the control points and measured points
	npts = this->importPointCSV(ctrlpts, false, QString("Select a control point file."));
	npts = this->importPointCSV(measpts, true, QString("Select a measure point file."));

	// disable the Start button if imports were not successful
	if( npts == 0)
		this->pushbutton_Start->setDisabled(true);
}

MainWindow::~MainWindow()
{
}

int MainWindow::importPointCSV(p3d* &p, bool doNptsCheck, QString title)
{
	// initialize variables
	QStringList filepath;	//string list populated with the file path by the dialog box
	QFile* readf = 0;		//imported file object
	QString templine;		//string to temporarily hold a line of imported text
	p = 0;					//the point structure being written to
	int n = 0;				//number of points being imported
	int i = 0;				//generic counter (used while populating point structure)
	bool typeIsOk = true;	//indicates whether a type error occurred while importing


	// select the file to import
	QFileDialog* fileDialog_readf = new QFileDialog(this, title);
		fileDialog_readf->setNameFilter("CSV (*.csv *.txt)");
		fileDialog_readf->setViewMode(QFileDialog::Detail);
		fileDialog_readf->setDirectory("C:\\Temp\\");
		fileDialog_readf->setFileMode(QFileDialog::ExistingFile);
		fileDialog_readf->setAcceptMode(QFileDialog::AcceptOpen);
	if( fileDialog_readf->exec() ) //verify that the user selected a file
	{
		filepath = fileDialog_readf->selectedFiles();
		delete fileDialog_readf;
		readf = new QFile(filepath[0]);
	} else
	{
		this->userUpdate(tr("Error: You must select a file to continue."));
		return 0;
	}

	// verify that the file exists and opens properly
	if( !readf->exists() )
	{
		this->userUpdate("Error: File load unsuccessful.\nFile does not exist.");
		return 0;
	}
	if( !readf->open( QIODevice::ReadOnly | QIODevice::Text ) )
	{
		this->userUpdate( "Error: File failed to open." );
		readf->close();
		return 0;
	}

	// count the number of points
	while( !readf->atEnd() )
	{
		templine = readf->readLine().trimmed();
		if( templine.indexOf("#") == 0 )
		{
			//ignore comment lines
		}
		else if( !templine.isEmpty() )
			n++;
	}
	this->userUpdate( QString("Importing %1 points from %2...").arg(n).arg(readf->fileName()) );

	// check against the existing npts count if instructed
	if( doNptsCheck )
	{
		if( n != this->npts )
		{
			this->userUpdate(tr("Error: Point count does not match."));
			readf->close();
			return 0;
		}
	}

	// import the point data
	p = new p3d[n];					//creates the point structure with the appropriate size
	readf->seek(0);					//rewind the file
	while( !readf->atEnd() )
	{
		templine = readf->readLine().trimmed();
		if( templine.indexOf("#") == 0 )
		{
			//ignore comment lines
		} else if( !templine.isEmpty() )
		{
			p[i].x = templine.section( ',', 0, 0 ).toDouble(&typeIsOk);
			p[i].y = templine.section( ',', 1, 1 ).toDouble(&typeIsOk);
			p[i].z = templine.section( ',', 2, 2 ).toDouble(&typeIsOk);
			i++;
		}
		if( !typeIsOk )
		{
			this->userUpdate( QString("Error: import type cast failure at point %1.").arg(i) );
			readf->close();
			return 0;
		}
	}

	// confirm successful import
	userUpdate( QString("%1 points successfully imported.").arg(n) );

	readf->close();
	return n;
}

void MainWindow::pushbutton_Start_onClicked()
{
// disable the pushbutton to ensure that it's not clicked again
	this->pushbutton_Start->setDisabled(true);
// clear the info window
	this->text_ViewLog->setPlainText("");
// perform adjustment
	AdjustNonLinear rot;
	double rms = 0;
	double deg = 0;
	double min = 0;
	double sec = 0;
	int i = rot.adjust(this->npts, this->measpts, this->ctrlpts, this->transpts, this->trans_params, rms );
// output results
	this->userUpdateFormatted(QString("<b>Adjustment Results:</b>"));
	this->userUpdate( QString("\tRMS = %1").arg(rms) );
	this->userUpdate( QString("\tNumber of iterations = %1").arg(i) );
	this->userUpdate( QString("\tS = %1").arg(this->trans_params->s, 0, 'f', 9) );
	this->userUpdate( QString("\tTx = %1").arg(this->trans_params->tx, 0, 'f', 3) );
	this->userUpdate( QString("\tTy = %1").arg(this->trans_params->ty, 0, 'f', 3) );
	this->userUpdate( QString("\tTz = %1").arg(this->trans_params->tz, 0, 'f', 3) );
	this->radToDMS( this->trans_params->ax, deg, min, sec );
	this->userUpdate( QString("\tRx = %1 (%2%3'%4\")").arg(this->trans_params->ax, 0, 'f', 9).arg(deg, 0, 'f', 0).arg(min, 0, 'f', 0).arg(sec, 0, 'f', 2) );
	this->radToDMS( this->trans_params->ay, deg, min, sec );
	this->userUpdate( QString("\tRy = %1 (%2%3'%4\")").arg(this->trans_params->ay, 0, 'f', 9).arg(deg, 0, 'f', 0).arg(min, 0, 'f', 0).arg(sec, 0, 'f', 2) );
	this->radToDMS( this->trans_params->az, deg, min, sec );
	this->userUpdate( QString("\tRz = %1 (%2%3'%4\")").arg(this->trans_params->az, 0, 'f', 9).arg(deg, 0, 'f', 0).arg(min, 0, 'f', 0).arg(sec, 0, 'f', 2) );
	this->userUpdateFormatted("<b>Rotation Matrix:</b>");
	for( int j = 0; j < 3; j++ )
		this->userUpdate( QString("\t%1\t%2\t%3").arg(rot.RotMat[j][0]).arg(rot.RotMat[j][1]).arg(rot.RotMat[j][2]) );
	this->userUpdateFormatted("<b>New Points:</b>");
	for( int j = 0; j < this->npts; j++ )
		this->userUpdate(QString("\t%1\t%2\t%3")
						 .arg(this->transpts[j].x, 0, 'f', 3)
						 .arg(this->transpts[j].y, 0, 'f', 3)
						 .arg(this->transpts[j].z, 0, 'f', 3)
						 );
}

void MainWindow::userUpdate(QString t)
{
	text_ViewLog->appendPlainText(t);
//	emit this->userUpdated();
}
void MainWindow::userUpdate(QString t, bool clear)
{
	if( clear )
		text_ViewLog->setPlainText("");
	text_ViewLog->appendPlainText(t);
//	emit this->userUpdated();
}
void MainWindow::userUpdateFormatted(QString t)
{
	text_ViewLog->appendHtml(t);
//	emit this->userUpdated();
}
void MainWindow::userUpdateFormatted(QString t, bool clear)
{
	if( clear )
		text_ViewLog->setPlainText("");
	text_ViewLog->appendHtml(t);
//	emit this->userUpdated();
}

void MainWindow::radToDMS (double angleRad, double &d, double &m, double &s)
{
	d = m = s = 0;
	double decDeg = angleRad * 180 / PI;
	d = floor( decDeg );
	m = floor( (decDeg - d)*60 );
	s = (decDeg - d)*3600 - m*60;
}
