% adjusts using the Han method, then applies the adjustment to measured data
function adjustHan(ctrlPath,measPath)
% ctrlPath and measPath are file paths to the control and measurement CSV
% files, following standard Windows ('C:\...*.txt') path format

format longG;

cpts = csvread(ctrlPath);
mpts = csvread(measPath);
num_cpts = size( cpts, 1 );
num_mpts = size( mpts, 1 );

if( num_cpts ~= num_mpts )
    disp('Error: Number of control points does not match measured points.')
    return;
end

% compute scale approximate and generate Delta-X-s & Delta-X-prime
% (eq 4, 6, 7)
dx_num = 0;     %counter for the dimensions of DXs and DXp
scale = 0;
for i = 1:1:num_cpts
    for j = i:1:num_cpts
        if( j > i )
            dx_num = dx_num + 1;
            
            mtx = mpts(i,1) - mpts(j,1); 
            mty = mpts(i,2) - mpts(j,2);
            mtz = mpts(i,3) - mpts(j,3);
            DXs(dx_num,:) = [mtx mty mtz];
            
            ctx = cpts(i,1) - cpts(j,1);
            cty = cpts(i,2) - cpts(j,2);
            ctz = cpts(i,3) - cpts(j,3);
            DXp(dx_num,:) = [ctx cty ctz];
            
            scale = scale + sqrt(ctx*ctx+cty*cty+ctz*ctz) / sqrt(mtx*mtx+mty*mty+mtz*mtz);
        end
    end
end
scale = scale / dx_num;
DXs = DXs * scale;

% compute rotation matrix (eq 5)
R = DXp' * DXs / ( DXs' * DXs );

% compute rotation angles (eq 8, 9)
rx = atan2( -R(3,2), R(3,3) );
ry = atan2( R(3,1), sqrt( R(1,1)*R(1,1) + R(2,1)*R(2,1) ) );
rz = atan2( -R(2,1), R(1,1) );

% compute translation vector (eq 10)
translation = zeros(3,1);
for i = 1:1:num_cpts
    translation = translation + cpts(i,:)' - scale*R*mpts(i,:)';
end
translation = translation/num_cpts;

% recompute the rotation matrix using the derived rotation angles
c1 = cos(rx); c2 = cos(ry); c3 = cos(rz);
s1 = sin(rx); s2 = sin(ry); s3 = sin(rz);
Rcalc(1,1) = c2*c3;
Rcalc(1,2) = c1*s3+s1*s2*c3;
Rcalc(1,3) = s1*s3-c1*s2*c3;
Rcalc(2,1) = -c2*s3;
Rcalc(2,2) = c1*c3-s1*s2*s3;
Rcalc(2,3) = s1*c3+c1*s2*s3;
Rcalc(3,1) = s2;
Rcalc(3,2) = -s1*c2;
Rcalc(3,3) = c1*c2;

% recompute the translation vector
translation2 = zeros(3,1);
for i = 1:1:num_cpts
    translation2 = translation2 + cpts(i,:)' - scale*Rcalc*mpts(i,:)';
end
translation2 = translation2/num_cpts;

% display results
disp('Raw Rotation Matrix (R-hat):'); disp(R);
disp('Recomputed Rotation Matrix (R-calc):'); disp(Rcalc);
disp('Derived rotation around x (alpha) (rad,deg):'); disp([rx radtodeg(rx)]);
disp('Derived rotation around y (beta) (rad,deg):'); disp([ry radtodeg(ry)]);
disp('Derived rotation around z (gamma) (rad,deg):'); disp([rz radtodeg(rz)]);
disp('Translation vector:'); disp(translation);
disp('Updated translation vector:'); disp(translation2);
disp('Scale:'); disp(scale);

% transform the points
tpts = zeros(3,1);
for i = 1:1:num_cpts
    tpts(:,i) = scale*R*mpts(i,:)' + translation;
end
disp('Transformed points (R-hat):'); disp(tpts');

% alternatively...
tptsCalc = zeros(3,1);
for i = 1:1:num_cpts
    tptsCalc(:,i) = scale*Rcalc*mpts(i,:)' + translation2;
end
disp('or... (R-calc):'); disp(tptsCalc');

% compute RMSE for both transformations
res = tpts' - cpts;
res = res.^2;
rmse = sqrt(sum(sum(res,1),2)/num_cpts);

resCalc = tptsCalc' - cpts;
resCalc = resCalc.^2;
rmseCalc = sqrt(sum(sum(resCalc,1),2)/num_cpts);

disp('RSME R-hat, R-Calc (new T):'); disp([rmse; rmseCalc]);
end