/*#######################################################################
  Still to add:
  [x] Abiity to create a new project in GUI
  [x] Batch processing (argv lines)
		- Auto-Real-time analysis
  [ ]	- Invisible mode with auto-start
  [ ] QProgressBar for file download
  [x] QProgressBar for scan progress
		- Scan abort button (MEAS_STOP())
		- Base it on QTimer functionality
  [ ] Need to reduce/increase resolution of monitor scan?
	  Camera functionality
  [x]	- Auto-exposure mode
  [ ]	- HDR mode
  [x] Don't disable the QPlainTextEdit
  [x] Override the QAbstractScrollArea SizeHint()
  [ ] Add icons to buttons
  [x] Update the scan parameters when the radio button is clicked
  [x] Copy the config file to ScanController so that the user can change settings during a scan for the next one. QObjects don't support copying (memcpy or =), so if I want to copy the settings then I'd either have to find another way to implement the VZConfig class, or copy it one component at a time myself.
  [ ] Actively download the monitor file instead of waiting, then downloading it
  [ ] Display the previous scan graphically (using the PNG?) and allow user to define limits based on that.
  [ ] Run ScanController within the context of a QThread so that waits and pauses don't lock the program. also opens the possibility for multiple threads.

  #####################################################################*/

#include <QtGui/QApplication>
#include "mainwindow.h"
#include <QStringList>
#include <QAction>

int processArguments(QStringList args, MainWindow* w)
{
	int mode = 0;						//program operation mode (return value - see MainWindow documentation)
	QString s; char c;					//the next input string and its first character
	bool typeOkay = true;				//an error indicator in case the user messed up input

	int itemp = 0;						//temporary variables to protect against crashes from incorrect input
	double dtemp = 0;
	for( int i = 0; i < args.count(); i++ )
	{
		s = args.at(i);
		c = s.at(0).toLatin1();
//		w->infoUpdate(QString("Argument %1: %2").arg(i).arg(s));
		switch(c)
		{
			//Program operation mode
			case 'M':
			case 'm':
				mode = s.section("=", 1, 1).toInt(&typeOkay);
				if(!typeOkay)
				{
					w->infoUpdate(QString("Error: wrong data type with argument %1. Running in basic mode.").arg(i));
					return 0;
				}
				//see MainWindow documentation for description of the modes
				if( mode == 0 ) return 0;	//ignore other arguments and escape if basic mode requested
				if( mode < 0 || mode > 2 )
				{
					w->infoUpdate("Operation mode outside acceptable range. Running in basic mode.");
					return 0;
				}
				break;

			//Project name
			case 'N':
			case 'n':
				//note that project names cannot have spaces or equal signs in them or else they won't parse properly
				w->newConfig->setProjectName(s.section("=", 1, 1) );

				//Assume that the project is a new one. If the project already exists, then the scanner will load it
				//and create a new scanpos... UNLESS it's on the internal drive and a usb drive is plugged in.
				w->newConfig->setIsNewProject(true);
				break;

			//Atmospheric settings (order: temperature, pressure, relative humidity)
			case 'A':
			case 'a':
				s.remove("a=", Qt::CaseInsensitive);
				itemp = s.section(",", 0, 0).toInt(&typeOkay);
				if( typeOkay ) w->newConfig->setTemperature( itemp );
				itemp = s.section(",", 1, 1).toInt(&typeOkay);
				if( typeOkay ) w->newConfig->setPressure( itemp );
				itemp = s.section(",", 2, 2).toInt(&typeOkay);
				if( typeOkay ) w->newConfig->setHumidity( itemp );
				break;

			//Scan Pattern (order: vstart, vstop, vincr, hstart, hstop, hincr)
			case 'P':
			case 'p':
				s.remove("p=", Qt::CaseInsensitive);
				//decide whether user requested a custom pattern or a preset pattern
				if( s.indexOf(",") < 0 )		//user requested a preset pattern
				{
					if( s == QString("Default") || s == QString("default") || s == QString("DEFAULT") )
						w->newConfig->setPatternPreset(VZConfig::SP_DEFAULT);
					else if( s == QString("2Min") || s == QString("2min") || s == QString("2MIN") )
						w->newConfig->setPatternPreset(VZConfig::SP_FULLTWOMINUTE);
					else if( s == QString("5Min") || s == QString("5min") || s == QString("5MIN") )
						w->newConfig->setPatternPreset(VZConfig::SP_FULLFIVEMINUTE);
					else
						typeOkay = false;
				} else							//user provided pattern parameters
				{
					vzpat ptemp;
					dtemp = s.section(",", 0, 0).toDouble(&typeOkay);
					if( typeOkay ) ptemp.vstart = dtemp;
					dtemp = s.section(",", 1, 1).toDouble(&typeOkay);
					if( typeOkay ) ptemp.vstop = dtemp;
					dtemp = s.section(",", 2, 2).toDouble(&typeOkay);
					if( typeOkay ) ptemp.vincr = dtemp;
					dtemp = s.section(",", 3, 3).toDouble(&typeOkay);
					if( typeOkay ) ptemp.hstart = dtemp;
					dtemp = s.section(",", 4, 4).toDouble(&typeOkay);
					if( typeOkay ) ptemp.hstop = dtemp;
					dtemp = s.section(",", 5, 5).toDouble(&typeOkay);
					if( typeOkay ) ptemp.hincr = dtemp;
					w->newConfig->setImportPattern(ptemp);
				}
				break;

			//Image Capture Toggle
			case 'I':
			case 'i':
				itemp = s.section("=", 1, 1).toInt(&typeOkay);
				if( itemp == VZConfig::IA_NONE || itemp == VZConfig::IA_NORMAL || itemp == VZConfig::IA_HDR )
					w->newConfig->setImageMode(itemp);
				break;

			//Real-Time Mode Toggle
			case 'R':
			case 'r':
				itemp = s.section("=", 1, 1).toInt(&typeOkay);
				if( itemp == 0 || itemp == 1 )
					w->newConfig->setIsRealTime(itemp);

			//ignore unknown tags
			default:
				break;

		} //end switch (argument tags)

		//verify proper type conversion with each argument
		if(!typeOkay)
		{
			w->infoUpdate(QString("Error: Wrong data type with argument %1.").arg(i+1));
			typeOkay = true;
		}
	}//end for(each argument)
	w->infoUpdate("Processing complete!");

	return mode;
}

int main(int argc, char *argv[])
{
	//initialize the application - MUST ALWAYS COME FIRST!
    QApplication a(argc, argv);
	//initialize the gui - required to initialize variable storage
	MainWindow w;

	//process the arguments
	QStringList args = a.arguments();	//the list of strings input by the user
	if( args.count() > 1 )
	{
		w.setRunMode( processArguments(args, &w) );
		w.finalizeImport();
	}
	args.clear();

    w.show();
    return a.exec();
}
