#include "mainwindow.h"

#include <QAction>
#include <QApplication>
#include <QDesktopServices>
#include <QDir>
#include <QFormLayout>
#include <QPlainTextEdit>
#include <QPushButton>
#include <QHBoxLayout>
#include <QUrl>
#include <QVBoxLayout>
#include <QLabel>
#include <QComboBox>
#include <QStringList>
#include <QGroupBox>
#include <QButtonGroup>
#include <QRadioButton>
#include <QDoubleSpinBox>
#include <QSpinBox>
#include <QCheckBox>
#include <QTextStream>
#include <QDockWidget>
#include <QGridLayout>
#include <QProgressBar>
#include <QTimer>
#include <QStatusBar>

const float VMIN = 30.;
const float VMAX = 130.;
const float VIMAX = 0.098750;
const float VIMIN = 0.001;
const float HMIN = 0.;
const float HMAX = 720.;
const float HIMAX = 0.499;
const float HIMIN = 0.001;


MainWindow::MainWindow(QWidget *parent) :
	QMainWindow(parent)
{
	//default to basic mode
	this->runMode = 0;

	// set up the configuration file(s)
	newConfig = new VZConfig;

	// build the user interface
	this->setWindowTitle("VZ400 Driver");
	setWindowFlags( windowFlags() & ~Qt::WindowMaximizeButtonHint );		//disable the Windows maximize button
	centralWidget = new QWidget(this);										//the centralWidget MUST be created & assigned for a MainWindow to be displayed
	this->setCentralWidget(centralWidget);
	setupGUI();
	connect(this, SIGNAL(busy(bool)), this, SLOT(on_IsBusy(bool)), Qt::DirectConnection);

	// update the Scan Pattern input boxes to display current values
	refreshFormDisplay();
	on_bgroup_Pattern_buttonClicked(bgroup_Pattern_ManualPreset->checkedButton());
	pushb_Config_RetrieveList->setFocus();
	pushb_Config_RetrieveList->setDefault(true);

	// set up the scan controller
	vzcontrol = new ScanController(this);
	vzcontrol->VZ_ScannerAddress = "S9997386";
	connect(vzcontrol, SIGNAL(scanningStarted()), this, SLOT(timer_ScanStart()));
	connect(vzcontrol, SIGNAL(cameraStarted(int)), this, SLOT(timer_CamStart(int)));
	connect(vzcontrol, SIGNAL(sessionFinished(bool)), this, SLOT(on_SessionFinished(bool)));
}

MainWindow::~MainWindow()
{

}

void MainWindow::changeImageMode()
{
	if(!check_Images->isChecked())
	{
		newConfig->setImageMode(newConfig->IA_NONE);
	} else if(!check_HDR->isChecked())
	{
		newConfig->setImageMode(newConfig->IA_NORMAL);
	} else
	{
		newConfig->setImageMode(newConfig->IA_HDR);
	}
}

void MainWindow::finalizeImport()
{
	//clean the imported project name and set it as the selected project
	QString name = newConfig->projectName();
	name.append(" (imported)");
	combo_Config_ProjectList_options << name;
	combo_Config_ProjectList->addItems(combo_Config_ProjectList_options);
	combo_Config_ProjectList->setCurrentIndex(0);

	//add the imported pattern and set it as the current pattern
	combo_Pattern_PresetSelect->addItem("Imported Pattern");
	combo_Pattern_PresetSelect->setCurrentIndex(combo_Pattern_PresetSelect->count() - 1);

	//refresh the user form to display imported data
	refreshFormDisplay();
}

void MainWindow::infoUpdate(QString info)
{
	this->text_InfoLog->appendPlainText(info);
	this->text_InfoLog->repaint();
}

void MainWindow::on_IsBusy(bool busy)
{
	if(busy)
	{
		centralWidget->setEnabled(false);
		qApp->setOverrideCursor(Qt::WaitCursor);
	} else
	{
		centralWidget->setEnabled(true);
		qApp->restoreOverrideCursor();
	}
}

void MainWindow::on_floatingDock(bool isFloating)
{
	//Adjusts the window size to its minimum size if the user just undocked the info log
	//This keeps the window from looking strangely spaced out
	if( isFloating )
	{
		this->adjustSize();
		southDock->setWindowTitle("Info Log (doubleclick here to re-dock)");
	} else
	{
		southDock->setWindowTitle("Info Log");
	}
}

void MainWindow::on_bgroup_Pattern_buttonClicked(QAbstractButton *b)
{
	if( b == this->radio_PatternPreset )
	{
		this->spinbox_Pattern_VStart->setEnabled(false);
		this->spinbox_Pattern_VStop->setEnabled(false);
		this->spinbox_Pattern_VIncr->setEnabled(false);
		this->spinbox_Pattern_HStart->setEnabled(false);
		this->spinbox_Pattern_HStop->setEnabled(false);
		this->spinbox_Pattern_HIncr->setEnabled(false);
		this->combo_Pattern_PresetSelect->setEnabled(true);

		this->newConfig->setPatternPreset(this->combo_Pattern_PresetSelect->currentIndex());
		this->refreshFormDisplay();
	} else if( b == this->radio_PatternUser )
	{
		this->spinbox_Pattern_VStart->setEnabled(true);
		this->spinbox_Pattern_VStop->setEnabled(true);
		this->spinbox_Pattern_VIncr->setEnabled(true);
		this->spinbox_Pattern_HStart->setEnabled(true);
		this->spinbox_Pattern_HStop->setEnabled(true);
		this->spinbox_Pattern_HIncr->setEnabled(true);
		this->combo_Pattern_PresetSelect->setEnabled(false);
	}
}

void MainWindow::on_combo_PresetSelect_currentIndexChanged(int i)
{
	newConfig->setPatternPreset(i);
	refreshFormDisplay();
}

void MainWindow::on_combo_ProjectList_currentIndexChanged(QString t)
{
	// strip the source tags (e.g., " (usb)" ) and store the project name
	t = t.remove(QRegExp(" \\(.{0,256}$"));
	newConfig->setProjectName(t);
	// indicate that the user selected an existing project
	if(newConfig->isNewProject())
		newConfig->setIsNewProject(false);
}

void MainWindow::on_combo_ProjectList_editTextChanged(QString t)
{
	newConfig->setProjectName(t);
	if(!newConfig->isNewProject())
		newConfig->setIsNewProject(true);
}

void MainWindow::on_pushb_CalcScanTime_clicked()
{
	// computes the scan time given the selected measurement program
	float t = newConfig->calcScanTime(this->combo_Config_MeasProg->currentIndex());
	this->infoUpdate(QString("Scan will take approximately %1 seconds.").arg(int(t)));
}

void MainWindow::on_pushb_OpenLog_clicked()
{
	// open the log file where RiVLib records the conversation with the scanner - use for troubleshooting
	QDesktopServices::openUrl(QUrl("file:///c:/temp/logfile.txt"));
}

void MainWindow::on_pushb_Start_clicked()
{
	// clicking the Start button also triggers the action_ScanBegin signal,
	// which in turn triggers on_ScanBegin(), so nothing
	// needs to happen here for now.
}

void MainWindow::on_ScanBegin()
{
	emit busy(true);					//lock the central UI to prevent unintended changes
	pushb_Start->setDefault(false);
	pushb_Start->setEnabled(false);
	vzcontrol->scanStart();				//scan runtime errors are handled internally

	/* Allow user interaction with the UI now that config settings can safely change */
	emit busy(false);
//	pushb_Start->setEnabled(false);
}

void MainWindow::on_SessionFinished(bool success)
{
	if( success )
	{
		if( newConfig->isRealTime() )
			writeScan();
	}

	this->pushb_Start->setEnabled(true);
	pushb_OpenLog->setFocus();
	pushb_OpenLog->setDefault(true);
//	emit busy(false);		//unlock the central UI

//	this->statusBar()->clearMessage();
}

void MainWindow::on_action_RetrieveProjectList_triggered()
{
	emit busy(true);

	QDir dir;
	QStringList filters;
	filters << "*.riproject";
	dir.setNameFilters(filters);

	// reset the project list if it already exists
	if(!combo_Config_ProjectList_options.isEmpty())
	{
		combo_Config_ProjectList_options.clear();
		combo_Config_ProjectList->clear();
	}

	// retrieve the *.riproject names from the scanner and its USB drive, if one is plugged in
	if( check_Debug->isChecked() )
	{
		// use the sample stored on the local drive if in debug mode
		dir.setPath("C:/dev/programs/sampledata");
		combo_Config_ProjectList_options = dir.entryList();
		// strip the suffix and add a source tag
		combo_Config_ProjectList_options.replaceInStrings(".riproject"," (debug)");
	} else
	{
		// retrieve project names from the scanner's USB drive (note: only looks at the USB drive's root directory)
		dir.setPath(QString("//%1/usb").arg(vzcontrol->VZ_ScannerAddress));
		combo_Config_ProjectList_options = dir.entryList();
		// strip the suffix and add a source tag
		combo_Config_ProjectList_options.replaceInStrings(".riproject"," (usb)");

		// retrieve project names from the scanner's internal drive
		dir.setPath(QString("//%1/intern/projects").arg(vzcontrol->VZ_ScannerAddress));
		combo_Config_ProjectList_options.append(dir.entryList());
		// strip the suffix and add a source tag
		combo_Config_ProjectList_options.replaceInStrings(".riproject"," (internal)");
	}

	// verify that the network address was found (that the scanner is connected)
	if( !combo_Config_ProjectList_options.isEmpty() )
	{
		// display a list of projects for the user to select from
		this->combo_Config_ProjectList->addItems(combo_Config_ProjectList_options);
		this->combo_Config_ProjectList->setEnabled(true);

		// transfer control over to the start button
		if(this->pushb_Config_RetrieveList->isDefault())
		{
			this->pushb_Config_RetrieveList->setDefault(false);
			this->pushb_Start->setDefault(true);
			this->pushb_Start->setFocus();
		}

		this->infoUpdate(tr("Project list download successful!"));
	} else
	{
		this->infoUpdate(tr("Project list download error: Could not connect to scanner."));
	}

	emit busy(false);
}

void MainWindow::refreshFormDisplay()
{
	/* updates the forms to display what is actually stored */
	combo_Config_Storage->setCurrentIndex(newConfig->storageMedia());
	combo_Config_MeasProg->setCurrentIndex(newConfig->measurementProgram());
	spinbox_Config_Temp->setValue(newConfig->temperature());
	spinbox_Config_Pres->setValue(newConfig->pressure());
	spinbox_Config_RH->setValue(newConfig->humidity());
	spinbox_Pattern_VStart->setValue(newConfig->vstart());
	spinbox_Pattern_VStop->setValue(newConfig->vstop());
	spinbox_Pattern_VIncr->setValue(newConfig->vincr());
	spinbox_Pattern_HStart->setValue(newConfig->hstart());
	spinbox_Pattern_HStop->setValue(newConfig->hstop());
	spinbox_Pattern_HIncr->setValue(newConfig->hincr());
	check_Images->setChecked(newConfig->imageMode());
	check_HDR->setChecked(newConfig->imageMode());
	if(newConfig->imageMode() == VZConfig::IA_HDR)
		check_HDR->setEnabled(true);
	else
		check_HDR->setEnabled(false);
	check_RealTime->setChecked(newConfig->isRealTime());
}

void MainWindow::setRunMode(int m)
{
	this->runMode = m;
}

void MainWindow::setupGUI()
{
	// build action list (important for later if menus and/or toolbars are added)
	action_ScanBegin = new QAction(this);
	connect(this->action_ScanBegin, SIGNAL(triggered()), this, SLOT(on_ScanBegin()));
	action_RetrieveProjectList = new QAction(this);

	// build string lists for the comboboxes
	QStringList combo_Config_Storage_options;
	combo_Config_Storage_options
			<< "USB/Internal (auto)"
			<< "Internal"
			<< "None (for testing only)";
	QStringList combo_Config_MeasProg_options;
	combo_Config_MeasProg_options
			<< "High Speed"
			<< "Long Range";
	QStringList combo_Pattern_Preset_options;
	combo_Pattern_Preset_options
			<< "(Default)"
			<< "Full 2min Overview"
			<< "Full 5min Overview";

	// assemble the Central Widget
	layout_Main = new QVBoxLayout;
		layout_Config = new QFormLayout();
			layout_Config_Projects = new QHBoxLayout();
				combo_Config_ProjectList = new QComboBox();
//					combo_Config_ProjectList->setEnabled(false);
					combo_Config_ProjectList->setEditable(true);
					connect(combo_Config_ProjectList, SIGNAL(currentIndexChanged(QString)), this, SLOT(on_combo_ProjectList_currentIndexChanged(QString)));
					connect(combo_Config_ProjectList, SIGNAL(editTextChanged(QString)), this, SLOT(on_combo_ProjectList_editTextChanged(QString)));
				layout_Config_Projects->addWidget(combo_Config_ProjectList);
				pushb_Config_RetrieveList = new QPushButton(tr("Fetch"));
					pushb_Config_RetrieveList->setToolTip(tr("Retrieve the project list from the scanner.\nA connection to the scanner is required."));
					pushb_Config_RetrieveList->setFixedWidth(40);
					connect(pushb_Config_RetrieveList, SIGNAL(clicked()), this, SLOT(on_action_RetrieveProjectList_triggered()));
				layout_Config_Projects->addWidget(pushb_Config_RetrieveList);
				check_Debug = new QCheckBox(tr("&Debug"));
					check_Debug->setChecked(false);
					check_Debug->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Preferred);
				layout_Config_Projects->addWidget(check_Debug);
			layout_Config->addRow(tr("Project:"), layout_Config_Projects);
			combo_Config_Storage = new QComboBox();
				combo_Config_Storage->addItems(combo_Config_Storage_options);
				connect(this->combo_Config_Storage, SIGNAL(currentIndexChanged(int)), this->newConfig, SLOT(setStorageMedia(int)));
			layout_Config->addRow(tr("Storage Media:"), combo_Config_Storage);
			combo_Config_MeasProg = new QComboBox();
				combo_Config_MeasProg->addItems(combo_Config_MeasProg_options);
				connect(this->combo_Config_MeasProg, SIGNAL(currentIndexChanged(int)), this->newConfig, SLOT(setMeasurementProgram(int)));
			layout_Config->addRow(tr("Measurement Program:"), combo_Config_MeasProg);
			layout_Config_Atmos = new QHBoxLayout();
				label_Config_Temp = new QLabel(tr("Temp:"));
					label_Config_Temp->setToolTip(tr("Current temperature."));
				layout_Config_Atmos->addWidget(label_Config_Temp);
				spinbox_Config_Temp = new QSpinBox();
					spinbox_Config_Temp->setToolTip(tr("Current temperature."));
					spinbox_Config_Temp->setMinimum(-50);
					spinbox_Config_Temp->setMaximum(100);
//					spinbox_Config_Temp->setValue(newConfig->temperature());
					spinbox_Config_Temp->setSuffix(QString("C"));
					connect(this->spinbox_Config_Temp, SIGNAL(valueChanged(int)), this->newConfig, SLOT(setTemperature(int)));
				layout_Config_Atmos->addWidget(spinbox_Config_Temp);
				label_Config_Pres = new QLabel(tr("Pr:"));
					label_Config_Pres->setToolTip(tr("Current atmospheric pressure\n(at sea level)."));
				layout_Config_Atmos->addWidget(label_Config_Pres);
				spinbox_Config_Pres = new QSpinBox();
					spinbox_Config_Pres->setToolTip(tr("Current atmospheric pressure\n(at sea level)."));
					spinbox_Config_Pres->setMinimum(600);
					spinbox_Config_Pres->setMaximum(1200);
//					spinbox_Config_Pres->setValue(newConfig->pressure());
					spinbox_Config_Pres->setSuffix("mbar");
					connect(this->spinbox_Config_Pres, SIGNAL(valueChanged(int)), this->newConfig, SLOT(setPressure(int)));
				layout_Config_Atmos->addWidget(spinbox_Config_Pres);
				label_Config_RH = new QLabel(tr("RH:"));
					label_Config_RH->setToolTip(tr("Current relative humidity."));
				layout_Config_Atmos->addWidget(label_Config_RH);
				spinbox_Config_RH = new QSpinBox();
					spinbox_Config_RH->setToolTip(tr("Current relative humidity."));
					spinbox_Config_RH->setMinimum(0);
					spinbox_Config_RH->setMaximum(100);
//					spinbox_Config_RH->setValue(newConfig->humidity());
					spinbox_Config_RH->setSuffix("%");
					connect(this->spinbox_Config_RH, SIGNAL(valueChanged(int)), this->newConfig, SLOT(setHumidity(int)));
				layout_Config_Atmos->addWidget(spinbox_Config_RH);
			layout_Config->addRow(tr("Atmospheric Conditions:"), layout_Config_Atmos);
		layout_Main->addLayout(layout_Config);

		groupbox_Pattern = new QGroupBox("Scan Pattern settings.");
			layout_Pattern = new QFormLayout();
				bgroup_Pattern_ManualPreset = new QButtonGroup();
					radio_PatternPreset = new QRadioButton(tr("Use preset."));
						radio_PatternPreset->setChecked(true);
					bgroup_Pattern_ManualPreset->addButton(radio_PatternPreset);
					radio_PatternUser = new QRadioButton(tr("Manually enter."));
					bgroup_Pattern_ManualPreset->addButton(radio_PatternUser);
					connect(this->bgroup_Pattern_ManualPreset, SIGNAL(buttonClicked(QAbstractButton*)), this, SLOT(on_bgroup_Pattern_buttonClicked(QAbstractButton*)));
				combo_Pattern_PresetSelect = new QComboBox();
					combo_Pattern_PresetSelect->addItems(combo_Pattern_Preset_options);
					connect(this->combo_Pattern_PresetSelect, SIGNAL(currentIndexChanged(int)), this, SLOT(on_combo_PresetSelect_currentIndexChanged(int)));
				layout_Pattern->addRow(radio_PatternPreset, combo_Pattern_PresetSelect);
				layout_Pattern->addRow(radio_PatternUser);

				layout_Pattern_Details = new QGridLayout();
					label_Pattern_Vert = new QLabel("Vertical");
					layout_Pattern_Details->addWidget(label_Pattern_Vert, 0, 0);
					layout_Pattern_Details->setAlignment(label_Pattern_Vert, Qt::AlignRight);
					label_Pattern_VStart = new QLabel("From:");
					layout_Pattern_Details->addWidget(label_Pattern_VStart, 0, 1);
					layout_Pattern_Details->setAlignment(label_Pattern_VStart, Qt::AlignHCenter);
					spinbox_Pattern_VStart = new QDoubleSpinBox();
						spinbox_Pattern_VStart->setMinimum(VMIN);
						spinbox_Pattern_VStart->setMaximum(VMAX);
						spinbox_Pattern_VStart->setSingleStep(5.0);
						spinbox_Pattern_VStart->setDecimals(3);
						connect(this->spinbox_Pattern_VStart, SIGNAL(valueChanged(double)), this->newConfig, SLOT(setVStart(double)));
					layout_Pattern_Details->addWidget(spinbox_Pattern_VStart, 0, 2);
					label_Pattern_VStop = new QLabel("To:");
					layout_Pattern_Details->addWidget(label_Pattern_VStop, 0, 3);
					layout_Pattern_Details->setAlignment(label_Pattern_VStop, Qt::AlignHCenter);
					spinbox_Pattern_VStop = new QDoubleSpinBox();
						spinbox_Pattern_VStop->setMinimum(VMIN);
						spinbox_Pattern_VStop->setMaximum(VMAX);
						spinbox_Pattern_VStop->setSingleStep(5.0);
						spinbox_Pattern_VStop->setDecimals(3);
						connect(this->spinbox_Pattern_VStop, SIGNAL(valueChanged(double)), this->newConfig, SLOT(setVStop(double)));
					layout_Pattern_Details->addWidget(spinbox_Pattern_VStop, 0, 4);
					label_Pattern_VIncr = new QLabel("Res:");
					layout_Pattern_Details->addWidget(label_Pattern_VIncr, 0, 5);
					layout_Pattern_Details->setAlignment(label_Pattern_VIncr, Qt::AlignHCenter);
					spinbox_Pattern_VIncr = new QDoubleSpinBox();
						spinbox_Pattern_VIncr->setMinimum(VIMIN);
						spinbox_Pattern_VIncr->setMaximum(VIMAX);
						spinbox_Pattern_VIncr->setSingleStep(0.001);
						spinbox_Pattern_VIncr->setDecimals(5);
						connect(this->spinbox_Pattern_VIncr, SIGNAL(valueChanged(double)), this->newConfig, SLOT(setVIncr(double)));
					layout_Pattern_Details->addWidget(spinbox_Pattern_VIncr, 0, 6);
					label_Pattern_Horiz = new QLabel("Horizontal");
					layout_Pattern_Details->addWidget(label_Pattern_Horiz, 1, 0);
					layout_Pattern_Details->setAlignment(label_Pattern_Horiz, Qt::AlignRight);
					label_Pattern_HStart = new QLabel("From:");
					layout_Pattern_Details->addWidget(label_Pattern_HStart, 1, 1);
					layout_Pattern_Details->setAlignment(label_Pattern_HStart, Qt::AlignHCenter);
					spinbox_Pattern_HStart = new QDoubleSpinBox();
						spinbox_Pattern_HStart->setMinimum(HMIN);
						spinbox_Pattern_HStart->setMaximum(360.0);
						spinbox_Pattern_HStart->setSingleStep(5.0);
						spinbox_Pattern_HStart->setDecimals(3);
						connect(this->spinbox_Pattern_HStart, SIGNAL(valueChanged(double)), this->newConfig, SLOT(setHStart(double)));
					layout_Pattern_Details->addWidget(spinbox_Pattern_HStart, 1, 2);
					label_Pattern_HStop = new QLabel("To:");
					layout_Pattern_Details->addWidget(label_Pattern_HStop, 1, 3);
					layout_Pattern_Details->setAlignment(label_Pattern_HStop, Qt::AlignHCenter);
					spinbox_Pattern_HStop = new QDoubleSpinBox();
						spinbox_Pattern_HStop->setMinimum(HMIN);
						spinbox_Pattern_HStop->setMaximum(HMAX);
						spinbox_Pattern_HStop->setSingleStep(5.0);
						spinbox_Pattern_HStop->setDecimals(3);
						connect(this->spinbox_Pattern_HStop, SIGNAL(valueChanged(double)), this->newConfig, SLOT(setHStop(double)));
					layout_Pattern_Details->addWidget(spinbox_Pattern_HStop, 1, 4);
					label_Pattern_HIncr = new QLabel("Res:");
					layout_Pattern_Details->addWidget(label_Pattern_HIncr, 1, 5);
					layout_Pattern_Details->setAlignment(label_Pattern_HIncr, Qt::AlignHCenter);
					spinbox_Pattern_HIncr = new QDoubleSpinBox();
						spinbox_Pattern_HIncr->setMinimum(HIMIN);
						spinbox_Pattern_HIncr->setMaximum(HIMAX);
						spinbox_Pattern_HIncr->setSingleStep(0.001);
						spinbox_Pattern_HIncr->setDecimals(5);
						connect(this->spinbox_Pattern_HIncr, SIGNAL(valueChanged(double)), this->newConfig, SLOT(setHIncr(double)));
					layout_Pattern_Details->addWidget(spinbox_Pattern_HIncr, 1, 6);
					layout_Pattern_Details->setColumnMinimumWidth(0, 70);
				layout_Pattern->addRow(layout_Pattern_Details);
			groupbox_Pattern->setLayout(layout_Pattern);
		layout_Main->addWidget(groupbox_Pattern);

		layout_Options = new QHBoxLayout();
			check_Images = new QCheckBox(tr("Capture Images"));
				check_Images->setToolTip(tr("Checking this will force image capture after scanning.\nThis lengthens scan time required significantly."));
				connect(check_Images, SIGNAL(clicked()), this, SLOT(changeImageMode()));
			layout_Options->addWidget(check_Images);
			check_HDR = new QCheckBox(tr("HDR Images"));
				check_HDR->setToolTip(tr("Checking this will record HDR images instead of normal images.\nThis takes long than normal images."));
				connect(check_HDR, SIGNAL(clicked()), this, SLOT(changeImageMode()));
				connect(check_Images, SIGNAL(clicked(bool)), check_HDR, SLOT(setEnabled(bool)));
			layout_Options->addWidget(check_HDR);
			check_RealTime = new QCheckBox(tr("Real-Time Mode"));
				check_RealTime->setToolTip(tr("Checking this will download the monitor RXP file\nand output an informational text file to be parsed\nfor pseudo-realtime analysis."));
				connect(check_RealTime, SIGNAL(clicked(bool)), this->newConfig, SLOT(setIsRealTime(bool)));
			layout_Options->addWidget(check_RealTime);
		layout_Main->addLayout(layout_Options);

		layout_Buttons = new QHBoxLayout;
			pushb_Start = new QPushButton(tr("Begin Scan"));
				pushb_Start->setToolTip(tr("Saves the current configuration and sends\nit to the scanner to begin a new scan.\nA connection to the scanner is required."));
				pushb_Start->setFixedWidth(80);
				connect(this->pushb_Start, SIGNAL(clicked()), this, SLOT(on_pushb_Start_clicked()));
				connect(this->pushb_Start, SIGNAL(clicked()), this->action_ScanBegin, SIGNAL(triggered()));
			layout_Buttons->addWidget(pushb_Start);
			pushb_CalcScanTime = new QPushButton(tr("Calc Duration"));
				pushb_CalcScanTime->setToolTip(tr("Estimates the scan duration in seconds."));
				pushb_CalcScanTime->setFixedWidth(80);
				connect(this->pushb_CalcScanTime, SIGNAL(clicked()), this, SLOT(on_pushb_CalcScanTime_clicked()));
			layout_Buttons->addWidget(pushb_CalcScanTime);
			pushb_OpenLog = new QPushButton(tr("Open Logfile"));
				pushb_OpenLog->setToolTip(tr("Opens the communication log with the scanner,\nprimarily for troubleshooting."));
				pushb_OpenLog->setFixedWidth(80);
				connect(this->pushb_OpenLog, SIGNAL(clicked()), this, SLOT(on_pushb_OpenLog_clicked()));
			layout_Buttons->addWidget(pushb_OpenLog);
		layout_Main->addLayout(layout_Buttons);

	centralWidget->setLayout(layout_Main);

	/* assemble the dock widget(s) */
	southDock = new QDockWidget();
		southDock->setAllowedAreas(Qt::BottomDockWidgetArea | Qt::TopDockWidgetArea);
		southDock->setFeatures(QDockWidget::NoDockWidgetFeatures | QDockWidget::DockWidgetFloatable);
		this->on_floatingDock(false);

		text_InfoLog = new MyPlainTextEdit();
			text_InfoLog->setMinimumSize(200,160);								//sets the minimimum (and default) widget width,height (optional)
			text_InfoLog->setReadOnly(true);									//prevents editing by the user (required when being used as a display widget)
			text_InfoLog->setTabChangesFocus(true);								//allows user to tab-cycle through the objects (Tab does nothing if false and it's ReadOnly)
			text_InfoLog->setLineWrapMode(QPlainTextEdit::NoWrap);				//prevents wordwrap so that each line of text is a unique message (optional)
//			text_InfoLog->setPlainText("Hello");								//sample starter text (optional. note that keeping this blank will clear the display)
//			text_InfoLog->appendHtml("<b>World</b>");							//sample addition of html-formatted text
//			text_InfoLog->setFont(QFont("Courier"));							//sample font change
//			text_InfoLog->setBackgroundVisible(true);
		southDock->setWidget(text_InfoLog);
		connect(southDock, SIGNAL(topLevelChanged(bool)), this, SLOT(on_floatingDock(bool)));
	this->addDockWidget(Qt::BottomDockWidgetArea, southDock);

	/* assemble the status bar */
	statusBar();
	pbar_ScanProgress = new QProgressBar();
		pbar_ScanProgress->setToolTip(tr("Current scan progress."));
		pbar_ScanProgress->setMinimum(0);
		pbar_ScanProgress->setMaximum(500);
		pbar_ScanProgress->setValue(0);
	pushb_ScanStop = new QPushButton("Stop");
		pushb_ScanStop->setToolTip(tr("Aborts the current scan\nand finishes downloading\nany queued data."));
		pushb_ScanStop->setFixedWidth(40);
	scanTimer = new QTimer(this);


	return;
}

void MainWindow::timer_ScanStart()
{
	pbar_ScanProgress->setValue(pbar_ScanProgress->minimum());
	pbar_ScanProgress->setMaximum(this->vzcontrol->VZ_TimeEstimate * 2 +9);	//multiply by 2 because the progress bar updates every half second, and add extra for the buffer
	pbar_ScanProgress->setFormat("%p%");

	statusBar()->addPermanentWidget(pbar_ScanProgress);
	statusBar()->addPermanentWidget(pushb_ScanStop);
	statusBar()->showMessage("Scanning in progress...");

	pbar_ScanProgress->show();
	pushb_ScanStop->show();

	pbar_ScanProgress->setEnabled(true);
	pushb_ScanStop->setEnabled(true);

	connect(scanTimer, SIGNAL(timeout()), this, SLOT(timer_ScanContinue()));
	connect(pushb_ScanStop, SIGNAL(clicked()), this, SLOT(timer_ScanCancelled()));

	scanTimer->start(500);			//update the progress bar every 500 milliseconds
}

void MainWindow::timer_ScanContinue()
{
	/* advance the progress bar forward */
	int step = pbar_ScanProgress->value()+1;
	if(step != pbar_ScanProgress->maximum())
	{
		pbar_ScanProgress->setValue(step);
	} else
	{
		/* stall at 99% to wait for the scanner to finish downloading from the buffer */
		if(!vzcontrol->checkBusy(false))
		{
			pbar_ScanProgress->setValue(step);
			infoUpdate(QString("Scan complete: %1").arg(vzcontrol->VZ_ScanName));
			timer_ScanFinished(true);
		}
	}
}

void MainWindow::timer_ScanCancelled()
{
	this->infoUpdate(tr("Scan aborted by user!"));
	this->timer_ScanFinished(false);
}

void MainWindow::timer_ScanFinished(bool success)
{
	scanTimer->stop();
	disconnect(scanTimer, SIGNAL(timeout()), 0, 0);
	disconnect(pushb_ScanStop, SIGNAL(clicked()), 0, 0);

	pbar_ScanProgress->setEnabled(false);
	pushb_ScanStop->setEnabled(false);
	statusBar()->showMessage("Scanning completed.");

	vzcontrol->scanFinalize(success);
}

void MainWindow::timer_CamStart(int numPhotos)
{
	/* the camera timer should only be enacted for normal image series acquisition */
	pbar_ScanProgress->setValue(1);
	pbar_ScanProgress->setMaximum(numPhotos);	//multiply by 2 because the progress bar updates every half second
	pbar_ScanProgress->setFormat("%v of %m");

//	this->statusBar()->addPermanentWidget(pbar_ScanProgress);
//	this->statusBar()->addPermanentWidget(pushb_ScanStop);
	statusBar()->showMessage("Image capture in progress...");
	pbar_ScanProgress->setEnabled(true);
	pushb_ScanStop->setEnabled(true);

	connect(scanTimer, SIGNAL(timeout()), this, SLOT(timer_CamContinue()));
	connect(pushb_ScanStop, SIGNAL(clicked()), this, SLOT(timer_CamCancelled()));

	scanTimer->start(1000);			//check every 1000 milliseconds to see if a new photo has been captured
}

void MainWindow::timer_CamContinue()
{
	/* check to see if a new photo has been acquired yet */
	int imagesCaptured = vzcontrol->camCheck();
	if(pbar_ScanProgress->value() == pbar_ScanProgress->maximum())
	{
		/* if we're on the last picture, quit if the camera has finished */
		if(!vzcontrol->checkBusy(false))
		{
			infoUpdate(tr("Image capture complete!"));
			timer_CamFinished(true);
		}
	} else
		if(imagesCaptured+1 > pbar_ScanProgress->value())
		{
			pbar_ScanProgress->setValue(imagesCaptured+1);
		}
}

void MainWindow::timer_CamCancelled()
{
	infoUpdate(tr("Image capture aborted by user!"));
	timer_CamFinished(false);
}

void MainWindow::timer_CamFinished(bool success)
{
	scanTimer->stop();
	disconnect(scanTimer, SIGNAL(timeout()), 0, 0);
	disconnect(pushb_ScanStop, SIGNAL(clicked()), 0, 0);

	pbar_ScanProgress->setEnabled(false);
	pushb_ScanStop->setEnabled(false);

	vzcontrol->camFinalize(success);
}

void MainWindow::writeScan()
{
	//this function assumes that the program is being run in real-time mode

	QString copyFrom(QString("//%1/%2.mon.rxp")
					 .arg(vzcontrol->VZ_ScannerAddress)
					 .arg(vzcontrol->VZ_ScanPath));
	QString copyTo(QString("%1/%2.mon.rxp")
				   .arg(QDir::currentPath())
				   .arg(vzcontrol->VZ_ScanName));

	//download the monitor file to the current directory
	infoUpdate( QString("Downloading monitor file to %1...").arg(copyTo) );
	if( QFile::copy(copyFrom, copyTo) )
		infoUpdate(tr("Scan download successful."));
	else
		infoUpdate(tr("Error: Scan download failed. File already exists?"));

	//output the real-time change data
	QFile writef("rt_info.txt");
	if( !writef.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		infoUpdate(tr("Error: Output file failed to open."));
		return;
	}
	QTextStream out(&writef);
	out << "#VZ_PROJ" << "\n";
	out << newConfig->projectName() << "\n\n";
	out << "#VZ_PATH" << "\n";
	out << copyTo << "\n\n";
	out << "#VZ_INCL" << "\n";
	out << vzcontrol->VZ_Roll << "\n";
	out << vzcontrol->VZ_Pitch << "\n";

	writef.close();
}
