#include <QMessageBox>
#include <QString>
#include <QPlainTextEdit>
#include <QApplication>
#include <QStatusBar>

#include <QStringBuilder>
#define QT_USE_FAST_CONCATENATION

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <stdexcept>
#include <qmath.h>
//#include <riegl/ctrlifc.h>
//#include <riegl/scanifc.h>

#include "scancontroller.h"

#if defined(_MSC_VER)
#   include <memory>
#else
#   include <tr1/memory>
#endif

//using namespace ctrllib;
//using namespace scanlib;
using namespace std;
//using namespace std::tr1;


ScanController::ScanController( MainWindow* parent )
{
	//gain control of the mainwindow's widgets
	mainwin = parent;

	/* generate the list of possible exposure times */
	exposureList << 0.0001 << 0.0002 << 0.0003 << 0.0004 << 0.0005 << 0.0006 << 0.0008 << 0.0010
			<< 0.0012 << 0.0015 << 0.0020 << 0.0025 << 0.0031 << 0.0040 << 0.0050 << 0.0062 << 0.0080
			<< 0.0100 << 0.0125 << 0.0166 << 0.0200 << 0.0250 << 0.0333 << 0.0400 << 0.0500 << 0.0666
			<< 0.0769 << 0.1000 << 0.1250 << 0.1666 << 0.2000 << 0.2500 << 0.3333 << 0.4000 << 0.5000
			<< 0.6250 << 0.7692 << 1.0000 << 1.3000 << 1.6000 << 2.0000 << 2.5000 << 3.0000 << 4.0000
			<< 5.0000 << 6.0000 << 8.0000 << 10.000 << 13.000 << 15.000 << 20.000 << 25.000 << 30.000;
}

ScanController::~ScanController()
{
	//ensure the connection to the scanner is closed
	session.close();
}

bool ScanController::checkBusy(bool wait)
{
	bool isbusy = false;
	std::string value;					//value of a requested property
	std::string result;					//receives result of a called function
	std::vector<std::string> comment;	//optional multi-line messages from the instrument
	std::string status;					//receives instrument's current status between commands
	try
	{
		if(wait)
		{
			/* wait for the scanner to not be busy */
			session.execute_command("MEAS_BUSY", "1", &result, &status, &comment);
			isbusy = QString::fromStdString(result).toInt();
			if(!isbusy)
			{
				session.execute_command("CAM_BUSY", "1", &result, &status, &comment);
				isbusy = QString::fromStdString(result).toInt();
			}
		} else
		{
			/* simply return whether or not the scanner is busy */
			session.execute_command("MEAS_BUSY", "0", &result, &status, &comment);
			isbusy = QString::fromStdString(result).toInt();
			if(!isbusy)
			{
				session.execute_command("CAM_BUSY", "0", &result, &status, &comment);
				isbusy = QString::fromStdString(result).toInt();
			}
		}
		return isbusy;
	}
	catch( exception &e )
	{
		mainwin->infoUpdate(QLatin1Literal("An error has occurred: ") % e.what());
		return true;
	}
	catch( ... )
	{
		mainwin->infoUpdate(QLatin1String("An unknown error has occurred: unknown exception."));
		return true;
	}
}

int ScanController::camCheck()
{
	/* check and return the current photo number being taken
	   returns -1 if an error occurs
	*/
	std::string value;					//value of a requested property
	std::string result;					//receives result of a called function
	std::vector<std::string> comment;	//optional multi-line messages from the instrument
	std::string status;					//receives instrument's current status between commands
	try
	{
		session.get_property("CAM_IMG_COUNTER", value);
		return QString::fromStdString(value).toInt();
	}
	catch( exception &e )
	{
		mainwin->infoUpdate(QLatin1Literal("An error has occurred: ") % e.what());
		return -1;
	}
	catch( ... )
	{
		mainwin->infoUpdate(QLatin1String("An unknown error has occurred: unknown exception."));
		return -1;
	}
}

void ScanController::camStart()
{
	int numPhotos;						//number of photos required to color the scan area

	std::string value;					//value of a requested property
	std::string result;					//receives result of a called function
	std::vector<std::string> comment;	//optional multi-line messages from the instrument
	std::string status;					//receives instrument's current status between commands
	try
	{
		/* retrieve the number of images required to color the scan area */
		session.execute_command("CAM_IMG_CALCNUM", "", &result, &status, &comment);
		numPhotos = QString::fromStdString(result).toInt();

		/* set image overlap */
		float camOverlap = 0.2; //default to 20% overlap for now
		session.set_property("CAM_OVERLAP", QString("%1").arg(camOverlap).toStdString());
		mainwin->infoUpdate(QLatin1Literal("Camera overlap set to ") % QString("%1").arg(camOverlap*100.) % QLatin1Literal("%"));

		mainwin->infoUpdate(QLatin1String("Capturing images..."));
		if(config->imageMode() == VZConfig::IA_NORMAL)
		{
			/* acquire normal images */
			session.set_property("CAM_EXPOSURE_MODE", "\"AUTOMATIC\"");
			session.execute_command("CAM_START", "1", &result, &status, &comment);
			emit this->cameraStarted(numPhotos);
		} else if(config->imageMode() == VZConfig::IA_HDR)
		{
			/* acquire HDR images */

			float spacing = (this->hstopActual - this->hstartActual) / numPhotos;	//the horizontal angle between each capture position
			int setSize = 3;									//number of photos per set; assume 3 photos per set for now

			/* populate the exposure list. might need to do this at every position. but that's slower */
			int* hdrExposures = new int[setSize];			//[0] is the base exposure index, then the overexposure(s), then the underexposure(s)
			session.set_property("CAM_EXPOSURE_MODE", "\"AUTOMATIC\"");
			session.get_property("CAM_EXPOSURE_TIME", value);
			hdrExposures[0] = exposureList.indexOf(QString::fromStdString(value).toFloat());
			hdrExposures[1] = this->exposureIndexValidate(hdrExposures[0]+7);
			hdrExposures[2] = this->exposureIndexValidate(hdrExposures[0]-6);

			/* aperture also varies with autoexposure mode, so record and reset that, too */
			session.get_property("CAM_LENS_APERTURE", value);
			session.set_property("CAM_EXPOSURE_MODE", "\"MANUAL\"");
			session.set_property("CAM_LENS_APERTURE", value);

			/* loop through each image set and capture LDR images */
			for( int set = 0; set < numPhotos; set++ )
			{

				/* get into position for the next set */
				float pos = (0.5 + set)*spacing;				//start in the middle of one space and go from there
				session.execute_command("SCN_ALIGN_PHI", QString("%1, 0").arg(pos).toStdString(), &result, &status, &comment);


				/* capture the LDR images for HDR processing */
				for( int i = 0; i < setSize; i++ )
				{
					qApp->processEvents();
					mainwin->infoUpdate("Taking image ");
					session.set_property("CAM_EXPOSURE_TIME", QString("%1").arg(exposureList.at(hdrExposures[i])).toStdString());
					session.execute_command("CAM_CAPTURE", "1", &result, &status, &comment);
//					session.execute_command("CAM_BUSY", "1", &result, &status, &comment);

					//need to add a pause here. Sometimes the CAM_CAPTURE command doesn't register fast enough to trigger CAM_BUSY, which means it moves onto the next photo before this one is done.
					//I'm thinking that I can use the camera timer and have a 1 second timer running at the start of each time through this for() loop
				}
			}

			/* set the camera values back to normal */
			session.set_property("CAM_EXPOSURE_MODE", "\"AUTOMATIC\"");
			session.execute_command("SCN_PARK", "", &result, &status, &comment);

			this->camFinalize(true);
		}
	}
	catch( exception &e )
	{
		mainwin->infoUpdate(QLatin1Literal("An error has occurred: ") % e.what());
		this->sessionExit(false);
	}
	catch( ... )
	{
		mainwin->infoUpdate(QLatin1String("An unknown error has occurred: unknown exception."));
		this->sessionExit(false);
	}
}

void ScanController::camFinalize(bool success)
{
	std::string value;					//value of a requested property
	std::string result;					//receives result of a called function
	std::vector<std::string> comment;	//optional multi-line messages from the instrument
	std::string status;					//receives instrument's current status between commands

	try
	{
		if(!success)
		{
			/* abort the image acquisition in progress if requested to do so */
			session.execute_command("CAM_STOP", "", &result, &status, &comment);
			session.execute_command("SCN_PARK", "", &result, &status, &comment);
		}
		mainwin->statusBar()->showMessage("Image capture completed.");

		this->sessionExit(success);
	}
	catch( exception &e )
	{
		mainwin->infoUpdate(QLatin1Literal("An error has occurred: ") % e.what());
		this->sessionExit(success);
	}
	catch( ... )
	{
		mainwin->infoUpdate(QLatin1String("An unknown error has occurred: unknown exception."));
		this->sessionExit(success);
	}
}

int ScanController::exposureIndexValidate(int i)
{
	/* validate that the exposure selected is actually in the list of possible exposures */
	if(i >= exposureList.size())
	{
		i = exposureList.size()-1;
	} else if(i < 0)
	{
		i = 0;
	}
	return i;
}

void ScanController::scanStart( )
{
	/* gain access to the config settings */
	config = new VZConfig();
	mainwin->newConfig->copyTo(config);

	std::string value;					//value of a requested property
	std::string result;					//receives result of a called function
	std::vector<std::string> comment;	//optional multi-line messages from the instrument
	std::string status;					//receives instrument's current status between commands
	QString update;						//used to output complex status updates
	QRegExp parser("(\"|/)");			//specifies an indexer to seek either a " or a /
	VZ_Roll = VZ_Pitch = 50000;			//intialize roll and pitch to be an impossible value

	/* Begin recording the telnet conversation with the scanner. Open this file to
	   troubleshoot command syntax and verify instructions sent to scanner
	*/
	session.start_logging("c://temp//logfile.txt");

	/* Surround instrument operation with C++ error handling, used by the RiVLib class.
	   Faulty command syntax will throw an error handled by the appropriate "catch"
	   routine, which should then display an error message to the user.
	*/
	try
	{
		/* Function **execute_command** calls a method of the instrument.
		   (execute_command("f", "1,2,3", ...) executes "f(1,2,3)")
		   Note: string values must be enclosed in quotation marks ("")
		   and multiple arguments are separated using a comma (,)
		*/
		/* Function **get_property** retrieves the value of a property from the instrument.
		   The value is returned as a string and may (depending on the property) consist
		   of multiple values separated by , (comma). Look up the property in the instrument
		   manual for a detailed description of the property and its values.
		*/
		/* Function **set_property** sets the value of a property with USER_WRITE privileges.
		   Note that some properties require the use of a method (execute_command) to set. See
		   the instrument manual for user privileges for each property.
		*/
		/* Function **start_logging** records the telnet conversation between the program
		   and instrument that is produced by the RiVLib commands.
		*/

		/* Connect to the specified instrument stored in the class as VZ_ScannerAddress.
		   The connection string is expected in format:
				"<host_name_or_ip_address>:[<service_name_or_port_number>]"
		   Examples:
				"192.168.0.234", "192.168.0.234:20002"
				"S9991234", "S9991234:20002"
		   If port number is not specified, 20002 (configuration port) is assumed as default.
		*/
		//The config port has the option to read a series of commands from a script file and do everything at once
//		session.execute_command("EXEC", "C:\Temp\VZ-Script.txt", &result, &status, &comment);

		mainwin->infoUpdate(QLatin1Literal("Attempting connection to ") % this->VZ_ScannerAddress % QLatin1Literal("..."));
		session.open(this->VZ_ScannerAddress.toStdString());
		mainwin->infoUpdate(QLatin1Literal("Connection to ") % this->VZ_ScannerAddress % QLatin1Literal(" successful!"));

		/* verify that the connected instrument is the proper model */
		session.get_property("INST_IDENT", value);
		mainwin->infoUpdate(QLatin1Literal("Instrument is a ") % QString::fromStdString(value));
		if( value.compare( 1, 6, "VZ-400" ) != 0 )
			throw runtime_error("VZ-400 instrument required.");

		/* wait for the instrument to finish any operation already underway */
		mainwin->infoUpdate(QLatin1String("Ensuring the instrument isn't already busy..."));
		checkBusy(true);

		/* verify that a camera is connected if image capture is enabled */
		if( config->imageMode() != VZConfig::IA_NONE )
		{
			session.get_property("CAM_DETECT", value);
			int cam_detected = QString::fromStdString(value).toInt();
			/* prompt the user whether to continue if no camera was detected */
			if(!cam_detected)
			{
				emit mainwin->busy(false);
				int proceed = QMessageBox::warning(mainwin, mainwin->windowTitle(),
												   "No camera was detected.\n"
												   "Proceed with scan anyway?",
												   QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes );
				if( proceed == QMessageBox::Yes )
				{
					mainwin->infoUpdate(QLatin1String("No camera was detected. Scan will proceed without capturing images."));
					config->setImageMode(VZConfig::IA_NONE);
					emit mainwin->busy(true);
				} else if( proceed == QMessageBox::No )
				{
					throw runtime_error("No camera was detected. User aborted the operation.");
				}
			}
		}

		/* disable automatic image acquisition */
		session.set_property("CAM_AUTOSTART", "\"OFF\"");

		/* set atmospheric conditions */
		update = QString("%1").arg(config->temperature());
		session.set_property("ATMOS_TEMP", update.toStdString());
		mainwin->infoUpdate(QLatin1Literal("Atmospheric temperature set to ") % update % QLatin1Literal("C"));
		update = QString("%1").arg(config->pressure());
		session.set_property("ATMOS_PRESSURE_SL", update.toStdString());
		mainwin->infoUpdate(QLatin1Literal("Atmospheric pressure set to ") % update % QLatin1Literal("mbar"));
		update = QString("%1").arg(config->humidity());
		session.set_property("ATMOS_REL_HUMID", update.toStdString());
		mainwin->infoUpdate(QLatin1Literal("Relative humidity set to ") % update % QLatin1Literal("%"));

		/* set data storage media for loading/saving project info */
		session.set_property("STOR_MEDIA", config->storageMediaToStd());
		mainwin->infoUpdate(QLatin1Literal("Storage media set to ") % QString::fromStdString(config->storageMediaToStd()));

		/* set current measurement program */
		session.execute_command("MEAS_SET_PROG", config->measurementProgramToStd(), &result, &status, &comment);
		mainwin->infoUpdate(QLatin1Literal("Measurement Program is now ") % QString::fromStdString(config->measurementProgramToStd()));

		/* set scan parameters... */
//		session.execute_command("SCN_LOAD_PAT", "\"2minOverview\"", &result, &status, &comment);
//		mainwin->infoUpdate(QLatin1String("2minOverview scan pattern loaded"));
		session.execute_command("SCN_SET_RECT_FOV", QString("%1, %2, %3, %4, %5, %6")
								.arg(config->vstart())
								.arg(config->vstop())
								.arg(config->vincr())
								.arg(config->hstart())
								.arg(config->hstop())
								.arg(config->hincr())
								.toStdString(), &result, &status, &comment);
		mainwin->infoUpdate(QLatin1String("Scan Pattern set to the following:"));
		mainwin->infoUpdate(QString("...Vertical:\tFrom %1 to %2, by %3.")
						 .arg(QString::fromStdString(comment[2]).section(" = ", 1, 1))
						 .arg(QString::fromStdString(comment[3]).section(" = ", 1, 1))
						 .arg(QString::fromStdString(comment[4]).section(" = ", 1, 1)));
		this->hstartActual = QString::fromStdString(comment[5]).section(" = ", 1, 1).toFloat();
		this->hstopActual = QString::fromStdString(comment[6]).section(" = ", 1, 1).toFloat();
		mainwin->infoUpdate(QString("...Horizontal:\tFrom %1 to %2, by %3.")
						 .arg(this->hstartActual)
						 .arg(this->hstopActual)
						 .arg(QString::fromStdString(comment[7]).section(" = ", 1, 1)));
		this->VZ_TimeEstimate = qCeil(QString::fromStdString(comment[18]).section(" = ", 1, 1).toFloat());
		mainwin->infoUpdate(QString("Scan will take approximately %1 seconds.")
						 .arg(this->VZ_TimeEstimate));

		/* create a new scanposition in the current project
		   note: leaving the STOR_SCANPOS_CREATE input string blank has the scanner automatically
		   assign a number to the ScanPosNNN folder
		*/
		if(config->isNewProject())
		{
			session.execute_command("STOR_PROJECT_CREATE", config->projectNameToStd(), &result, &status, &comment);
			mainwin->infoUpdate(QLatin1Literal("...STOR_PROJECT_CREATE returned: ") % QString::fromStdString(result));
			config->setIsNewProject(false);
		} else
		{
			session.execute_command("STOR_PROJECT_LOAD", config->projectNameToStd(), &result, &status, &comment);
			mainwin->infoUpdate(QString("Loaded project %1 from %2 drive.")
							 .arg(config->projectName())
							 .arg(QString::fromStdString(result).section(parser, 1, 1)));
		}
		session.execute_command("STOR_SCANPOS_CREATE", "\"\"", &result, &status, &comment);
		mainwin->infoUpdate(QLatin1Literal("Scan position ") % QString::fromStdString(result).right(12).left(10) % QLatin1Literal(" created."));

		/* measure inclination data for real-time analysis */
		if(config->isRealTime())
		{
			mainwin->infoUpdate(QLatin1String("Measuring inclination data..."));
			session.execute_command("INCL_ACQUIRE", "\"AVERAGE_WITH_CONTINUOUS_ROTATION\"", &result, &status, &comment);
			update = QString::fromStdString(result);
			this->VZ_Roll = update.section(",", 0, 0).toFloat();
			this->VZ_Pitch = update.section(",", 1, 1).toFloat();
			mainwin->infoUpdate(QString("Roll,Pitch = %1,%2")
							 .arg(this->VZ_Roll)
							 .arg(this->VZ_Pitch));
		}

		/* begin scanning - the program leaves the data on the scanner's media for now */
		mainwin->infoUpdate(QLatin1String("Scanning in progress..."));
		session.execute_command("MEAS_START", "", &result, &status, &comment);
		this->VZ_ScanPath = QString::fromStdString(result).remove('\"');
		this->VZ_ScanName = this->VZ_ScanPath.right(13);

		emit this->scanningStarted();
	}
	catch( exception &e )
	{
		/* inform the user of an error having occurred. */
		mainwin->infoUpdate(QLatin1Literal("An error has occurred: ") % e.what());
		this->sessionExit(false);
//		this->scanFinalize(false);
	}
	catch( ... )
	{
		/* this shouldn't happen, but just in case this will manage unknown error
		   messages thrown by RiVLib
		*/
		mainwin->infoUpdate(QLatin1String("An unknown error has occurred: unknown exception."));
		this->sessionExit(false);
	}
}

void ScanController::scanFinalize(bool success)
{
	std::string value;					//value of a requested property
	std::string result;					//receives result of a called function
	std::vector<std::string> comment;	//optional multi-line messages from the instrument
	std::string status;					//receives instrument's current status between commands

	try
	{
		if(!success)
		{
			/* abort the scan in progress if requested to do so */
			session.execute_command("MEAS_STOP", "", &result, &status, &comment);

			/* park the scanner if a scan failed for any reason */
			session.execute_command("SCN_PARK", "", &result, &status, &comment);
		}
		this->checkBusy(true);

		/* re-enable automatic image acquisition */
		session.set_property("CAM_AUTOSTART", "\"AUTO\"");

		/* acquire images if desired */
		if(config->imageMode() && success)
			this->camStart();
		else
			this->sessionExit(success);
	}
	catch( exception &e )
	{
		mainwin->infoUpdate(QLatin1Literal("An error has occurred: ") % e.what());
		this->sessionExit(success);
	}
	catch( ... )
	{
		mainwin->infoUpdate(QLatin1String("An unknown error has occurred: unknown exception."));
		this->sessionExit(success);
	}
}

void ScanController::sessionExit(bool success)
{
	/* end the session */
	delete config;
	session.stop_logging();	//close communication logging
	session.close();		//close the connection to the scanner
	emit this->sessionFinished(success);
}
