#include "vzconfig.h"

const float VMIN = 30.;
const float VMAX = 130.;
const float VIMAX = 0.098750;
const float VIMIN = 0.001;
const float HMIN = 0.;
const float HMAX = 720.;
const float HIMAX = 0.499;
const float HIMIN = 0.001;

VZConfig::VZConfig()
{
	// set defaults
	this->restoreDefaults();
}

//	############################################
//	#####  Convenience Functions  ##############
//	############################################

float VZConfig::calcScanTime(int mp)
{
	float pulset;				//length of time for one pulse
	float hdelta, vdelta;		//range of the horizontal, vertical angles
	float linet;				//length of time per vertical line

	hdelta = this->HStop - this->HStart;
	vdelta = this->VStop - this->VStart;

	switch(mp)
	{
	case MP_HIGHSPEED:
		pulset = 300000;	//high speed pulses are at 300000Hz
		break;
	case MP_LONGRANGE:
		pulset = 120000;	//long range pulses are at 120Hz
		break;
	default:
		pulset = 300000;
	}

	linet = (vdelta/this->VIncr) * 2.4 / pulset;

	return (hdelta/this->HIncr) * linet;
}

void VZConfig::copyTo(VZConfig *cfg)
{
	cfg->setProjectName(this->projectName());
	cfg->setTemperature(this->temperature());
	cfg->setPressure(this->pressure());
	cfg->setHumidity(this->humidity());
	cfg->setStorageMedia(this->storageMedia());
	cfg->setMeasurementProgram(this->measurementProgram());
	cfg->setIsNewProject(this->isNewProject());
	cfg->setImageMode(this->imageMode());
	cfg->setIsRealTime(this->isRealTime());

	cfg->setHStart(this->hstart());
	cfg->setHStop(this->hstop());
	cfg->setHIncr(this->hincr());
	cfg->setVStart(this->vstart());
	cfg->setVStop(this->vstop());
	cfg->setVIncr(this->vincr());
}

void VZConfig::restoreDefaults()
{
	setProjectName("");
	setTemperature(12);		// -50 - 100C	(int8)
	setPressure(1000);		// 0 - 1200mbar	(uint32)
	setHumidity(60);		// 0 - 100%		(uint8)
	setStorageMedia(SM_AUTO);
	setMeasurementProgram(MP_HIGHSPEED);
	setIsNewProject(true);
	setImageMode(IA_NONE);
	setIsRealTime(false);
	setPatternPreset(SP_DEFAULT);
}

//	############################################
//	#####  Data Retrieval Functions  ###########
//	############################################
int VZConfig::temperature()
{
	return AtmosTemperature;
}
int VZConfig::pressure()
{
	return AtmosPressure;
}
int VZConfig::humidity()
{
	return AtmosHumidity;
}

int VZConfig::measurementProgram()
{
	return MeasurementProgram;
}
std::string VZConfig::measurementProgramToStd()
{
	switch(MeasurementProgram)
	{
	case MP_HIGHSPEED:
		return "\"HIGH SPEED\"";
		break;
	case MP_LONGRANGE:
		return "\"LONG RANGE\"";
		break;
	default:
		return "\"ERROR: BAD MEASUREMENT PROGRAM SELECTION\"";
	}
}
int VZConfig::storageMedia()
{
	return StorageMedia;
}
std::string VZConfig::storageMediaToStd()
{
	switch(StorageMedia)
	{
	case SM_AUTO:
		return "\"AUTO USB/INT\"";
		break;
	case SM_INTERNAL:
		return "\"INTERN\"";
		break;
	case SM_NONE:
		return "\"NONE\"";
		break;
	default:
		return "\"ERROR: BAD STORAGE MEDIA SELECTION\"";
	}
}

float VZConfig::hincr()
{
	return HIncr;
}
float VZConfig::hstart()
{
	return HStart;
}
float VZConfig::hstop()
{
	return HStop;
}
float VZConfig::vincr()
{
	return VIncr;
}
float VZConfig::vstart()
{
	return VStart;
}
float VZConfig::vstop()
{
	return VStop;
}


int VZConfig::imageMode()
{
	return ImageMode;
}

bool VZConfig::isNewProject()
{
	return IsNewProject;
}
bool VZConfig::isRealTime()
{
	return IsRealTime;
}
QString VZConfig::projectName()
{
	return ProjectName;
}
std::string VZConfig::projectNameToStd()
{
	return QString("\"%1\"").arg(ProjectName).toStdString();
}

//	############################################
//	#####  Data Storage Slots  #################
//	############################################

void VZConfig::setTemperature(int val)
{
	AtmosTemperature = val;
}
void VZConfig::setPressure(int val)
{
	AtmosPressure = val;
}
void VZConfig::setHumidity(int val)
{
	AtmosHumidity = val;
}

void VZConfig::setMeasurementProgram(int val)
{
	MeasurementProgram = val;
}

void VZConfig::setStorageMedia(int val)
{
	StorageMedia = val;
}

void VZConfig::setPatternPreset(int p)
{
	switch(p)
	{
	case SP_DEFAULT:
		setVStart(30.);
		setVStop(130.);
		setVIncr(0.09875);
		setHStart(0.);
		setHStop(360.);
		setHIncr(0.4);
		break;
	case SP_FULLTWOMINUTE:
		setVStart(30.);
		setVStop(130.);
		setVIncr(0.05);
		setHStart(0.);
		setHStop(360.);
		setHIncr(0.05);
		break;
	case SP_FULLFIVEMINUTE:
		setVStart(30.);
		setVStop(130.);
		setVIncr(0.03);
		setHStart(0.);
		setHStop(360.);
		setHIncr(0.03);
		break;
	case SP_IMPORT:
		setVStart(importPattern.vstart);
		setVStop(importPattern.vstop);
		setVIncr(importPattern.vincr);
		setHStart(importPattern.hstart);
		setHStop(importPattern.hstop);
		setHIncr(importPattern.hincr);
		break;
	}
}
void VZConfig::setImportPattern(vzpat p)
{
	importPattern.vstart = p.vstart;
	importPattern.vstop = p.vstop;
	importPattern.vincr = p.vincr;
	importPattern.hstart = p.hstart;
	importPattern.hstop = p.hstop;
	importPattern.hincr = p.hincr;
}

void VZConfig::setHIncr(double val)
{
	if( val < HIMIN )
		HIncr = HIMIN;
	else if( val > HIMAX )
		HIncr = HIMAX;
	else
		HIncr = float(val);
}
void VZConfig::setHStart(double val)
{
	if( val < HMIN )
		HStart = HMIN;
	else if( val > HMAX )
		HStart = HMAX;
	else
		HStart = float(val);
}
void VZConfig::setHStop(double val)
{
	if( val < HMIN )
		HStop = HMIN;
	else if( val > HMAX )
		HStop = HMAX;
	else
		HStop = float(val);
}
void VZConfig::setVIncr(double val)
{
	if( val < VIMIN )
		VIncr = VIMIN;
	else if( val > VIMAX )
		VIncr = VIMAX;
	else
		VIncr = float(val);
}
void VZConfig::setVStart(double val)
{
	if( val < VMIN )
		VStart = VMIN;
	else if( val > VMAX )
		VStart = VMAX;
	else
		VStart = float(val);
}
void VZConfig::setVStop(double val)
{
	if( val < VMIN )
		VStop = VMIN;
	else if( val > VMAX )
		VStop = VMAX;
	else
		VStop = float(val);
}

void VZConfig::setImageMode(int val)
{
	ImageMode = val;
}

void VZConfig::setIsNewProject(bool val)
{
	IsNewProject = val;
}
void VZConfig::setIsRealTime(bool val)
{
	IsRealTime = val;
}
void VZConfig::setProjectName(QString qstr)
{
	ProjectName = qstr;
}
