/*********************************************************************************************
* This class serves as a container for storing the VZ-400's configuration settings in a format
* that the scanner and user can both understand. Used in combination with the ScanController
* class, it is meant to make downloading, storing, and uploading scanner settings simpler
* in an object-oriented programming context.
*
* Authored by Evon Silvia, School of Civil and Construction Engineering
* October 20, 2010
*********************************************************************************************/

#include <QObject>

#ifndef VZCONFIG_H
#define VZCONFIG_H

struct vzpat
{
	float vstart, vstop, vincr;			//vertical (theta) angle start, stop, and increment
	float hstart, hstop, hincr;			//horizontal (phi) angle start, stop, and increment
};

class VZConfig : public QObject
{
	Q_OBJECT

//	############################################
//	#####  VZ Config Functions  ################
//	############################################
public:
	VZConfig();

	//convenience functions
	float calcScanTime(int mp);
	void copyTo(VZConfig* cfg);
	void restoreDefaults();

	//data retrieval functions
	int temperature();
	int pressure();
	int humidity();

	int			measurementProgram();
	std::string	measurementProgramToStd();
	int			storageMedia();
	std::string	storageMediaToStd();

	float hincr();		// 0 - 0.499443	(0.000001 precision)
	float hstart();		// 0 - 360		(0.001 precision)
	float hstop();		// 0 - 720		(0.001 precision)
	float vincr();		// 0 - 0.098750	(0.000001 precision) 0.05 for 2min
	float vstart();		// 30- 130		(0.001 precision)
	float vstop();		// 30- 130		(0.001 precision)

	int imageMode();

	bool isNewProject();
	bool isRealTime();
	QString projectName();
	std::string projectNameToStd();


//	############################################
//	#####  VZ Config Slots  ####################
//	############################################
public slots:
	void setTemperature(int val);
	void setPressure(int val);
	void setHumidity(int val);

	void setMeasurementProgram(int val);
	void setStorageMedia(int val);

	void setPatternPreset(int p);
	void setImportPattern(vzpat p);
	void setHIncr(double val);
	void setHStart(double val);
	void setHStop(double val);
	void setVIncr(double val);
	void setVStart(double val);
	void setVStop(double val);

	void setImageMode(int val);

	void setIsNewProject(bool val);
	void setIsRealTime(bool val);
	void setProjectName(QString qstr);


//	############################################
//	#####  VZ Config Variables  ################
//	############################################
public:
	enum ScanPatternPresets
	{
		SP_DEFAULT,
		SP_FULLTWOMINUTE,
		SP_FULLFIVEMINUTE,
		SP_IMPORT
	};
	enum MeasProg
	{
		MP_HIGHSPEED, //Runs the scanner at its fastest standard mode (300Hz)
		MP_LONGRANGE  //Runs the scanner at its maximum range and power (120Hz)
	};
	enum StorMedia
	{
		SM_AUTO,		//selects USB if a drive exists there, internal if not
		SM_INTERNAL,	//forces the instrument to use the scanner's internal drive
		SM_NONE			//doesn't save the data stream to any location (can still connect to the data stream via RDTP)
	};
	enum ImgAcq
	{
		IA_NONE,		//no image acquisition
		IA_NORMAL,		//normal image series acquisition
		IA_HDR			//HDR image acquisition
	};

protected:
	int	AtmosTemperature;
	int	AtmosPressure;
	int	AtmosHumidity;

	int	MeasurementProgram;
	int StorageMedia;

	// project management
	bool IsNewProject;
	bool IsRealTime;
	QString	ProjectName;

	// scan pattern
	std::string ScanPatternName;		//string for holding preset scan pattern names (currently unused)
	float VStart, VStop, VIncr;			//vertical (theta) angle start, stop, and increment
	float HStart, HStop, HIncr;			//horizontal (phi) angle start, stop, and increment
	vzpat importPattern;

	// image acquisition settings
	int ImageMode;

//	vzpat* pat;							//scan pattern parameters
};

#endif // VZCONFIG_H
